# NHigh - version 2.0.3 (eru_bahagoon's BETA)
# http://metanet.2.forumer.com/index.php?showtopic=6697
# Written by jg9000  (jg9000 at gmail.com)
#
# You may freely modify and distribute this program, as long
# as I am credited.



'''Eru_bahagon has edited this programm.
Most edit are flaged. Visit the forum if you have any questions'''
# http://forum.ninjarobotyeti.com/viewtopic.php?f=20&t=9088

import urllib, cPickle, datetime, httplib, time, sys, os, re
import struct

NUM_EPISODES = 100
FRAMES_PER_SEC = 40
FRAME_TIME = 0.025



### The following is needed for python 2.3 support:   
if not hasattr(__builtins__,'set'):
    from sets import Set as set
def itemgetter(index): #same as operator.itemgetter
    return lambda x: x[index]
def sortList(lst, key=None, reverse=False):
    if sys.version_info >= (2,4):
        lst.sort(key=key, reverse=reverse)
    else:
        if key:
            lst2 = [(key(x),x) for x in lst]
            lst2.sort()
            lst[:] = [x[1] for x in lst2]
        else:
            lst.sort()
        if reverse:
            lst.reverse()
##################################################


####### config #######

class _ConfigField(object):
    def __init__(self, name, default, readFunc=str, writeFunc=str):
        self.name = name
        self.default = default
        self.readFunc = readFunc
        self.writeFunc = writeFunc

def _cfg_strToList(val):
    return val.split()
def _cfg_listToStr(val):
    return '  '.join(val)
class NHighConfig(object):
    def __init__(self):
        self.fields = [
            _ConfigField('defaultFilename', 'latest.hs'),
            _ConfigField('defaultPlayerName', self._getPlayerFromSol),
            _ConfigField('allowZerothTies', 1, readFunc=int),
            _ConfigField('ignoredPlayers', ['ANGERFIST',
                                            'crappitrash',
                                            'kryX-orange',
                                            'L3X',
                                            'naem',
                                            'Bonzai',
                                            'Goo',
                                            'Sp33dy',
                                            'pokemaniac1342',
                                            'ACEOFSPADEWINS',
                                            'Vegeta',
                                            'BAS3',
                                            'cheese_god',
                                            'fuckingyourdad',
                                            'sry4trbleIMdone',
                                            'haxYOscoreboard',
                                            'haxyoscoreboard',
                                            'igotbored:(',
                                            'Cuppy33'],
                         readFunc=_cfg_strToList, writeFunc=_cfg_listToStr),
            _ConfigField('nHighVersion', '2.0.3')
        ]
        self.fieldsDict = dict([(fld.name, fld) for fld in self.fields])
        self._findConfigFile()
        self.read()
    
    def read(self):
        for fld in self.fields:
            setattr(self, fld.name, fld.default)

        try:
            print "Config file location :",self.configFile
            f=file(self.configFile,'r')
            fdata = f.read()
            f.close()
        except (IOError,OSError):
            fdata = ''

        for line in fdata.splitlines():
            sp = [x.strip() for x in line.split('=',1)]
            if len(sp)!=2: continue
            fldname, fldval = sp
            if fldname not in self.fieldsDict: continue
            fld = self.fieldsDict[fldname]
            try:
                fldval = fld.readFunc(fldval)
            except (TypeError, ValueError):
                continue
            setattr(self, fld.name, fldval)
            
        for fld in self.fields:
            val = getattr(self, fld.name)
            if hasattr(val, '__call__'):
                #it's a function that will retun the actual value.
                setattr(self, fld.name, val())

        self.onUpdated()
        if not os.path.exists(self.configFile):
            try:
                self.save()
            except (IOError, OSError):
                pass
        
    def save(self): #doesn't catch IOError
        self.onUpdated()
        if not os.path.exists(self.configDir):
            os.makedirs(self.configDir)
        f = file(self.configFile, 'w')
        try:
            f.write('''\
# * NHigh configuration file *
# This is an automatically generated file.

''')
            for fld in self.fields:
                fldval = getattr(self, fld.name)
                fldval = fld.writeFunc(fldval)
                f.write('%s = %s\n\n' % (fld.name, fldval))
        finally:
            f.close()

    def onUpdated(self):
        #a set avoids duplicates
        self.ignoredPlayersSet = set([x.lower() for x in self.ignoredPlayers])

    def _findConfigFile(self):
        self.configDir = findPreferencesDir('NHigh')
        if not self.configDir:
            self.configDir = os.curdir
        self.configFile = os.path.join(self.configDir, 'nhighconfig')
        return self.configFile

    def _getPlayerFromSol(self):
        '''When the config file didn't exist, get the default
        player name from the .sol file'''
        try:
            solData = readSolFile()
            val = unicode2str(solData['username'])
            if val=='anon': #uninitialized
                return 'jg9000'
            return val
        except (IOError, OSError, NHighError, TypeError, KeyError):
            return 'jg9000'

#####################

def unicode2str(u, errors='replace'):
    return u.encode('latin1', errors)

class NHighError(StandardError):
    pass

class AppURLopener(urllib.FancyURLopener):
    def __init__(self, *args):
        self.version = "NHigh/2.0"
        urllib.FancyURLopener.__init__(self, *args)
urllib._urlopener = AppURLopener()

def openURL(url, data=None):
    try:
        f = urllib.urlopen(url, data)
        ret = f.read()
        f.close()
        return ret
    except AttributeError: #workaround a urllib bug
        raise IOError('HTTP error: connection reset')
    except httplib.HTTPException as e:
        raise IOError, ('http error: '+str(e)), sys.exc_info()[2]

class HighScore(object):
    def __init__(self, name=None, score=None):
        self.name = name
        self.score = score
    def gettime(self):
        return self.score * FRAME_TIME
    time = property(gettime)
    def __str__(self):
        return '[%s, %0.3f]'%(self.name,self.time)
    def __repr__(self):
        return str(self)
    def __eq__(self, other):
        return hasattr(other,'name') and self.name==other.name and \
               hasattr(other,'score') and self.score==other.score
    def __ne__(self, other):
        return not self.__eq__(other)
    def __hash__(self):
        return hash((self.name, self.score))

class HSTable(object):
    def __init__(self, table=None, timestamp=None):
        if timestamp is None:
            timestamp = datetime.datetime.now().replace(microsecond=0)
        self.timestamp = timestamp
        self.table = table

emptyHighScore = HighScore('', 0)

def _postProcessLevelScores(lvlScores):
    nameslist = [] #to get rid of double entries
    #for i in xrange(len(lvlScores)-1,-1,-1):
    for i in xrange(len(lvlScores)):
        entry = lvlScores[i]
        if entry.name.lower() in config.ignoredPlayersSet:
            del lvlScores[i]
            lvlScores.append(emptyHighScore)
        elif entry.name.lower() in nameslist:
            del lvlScores[i]
            lvlScores.append(emptyHighScore)
        else : 
            nameslist.append(entry.name.lower())

def _downloadAllEpisodeScoresOnce(ep, ret):
    url = 'http://www.harveycartel.org/metanet/n/data13/get_topscores_query_jg.php'
    postdata = 'episode%5Fnumber=' + str(ep)
    try:
        data = openURL(url, postdata).replace('\r','').split('&')
    except IOError:
        return False
    dct = {}
    for assignment in data:
        if not assignment: continue
        sp = assignment.split('=',1)
        if len(sp)!=2: continue
        dct[sp[0]] = sp[1]
    for lvl in xrange(6):
        lvlkey = '01234e'[lvl]
        try:
            for rank in xrange(20):
                scorekey = lvlkey+'score'+str(rank)
                namekey = lvlkey+'name'+str(rank)
                score = int(dct[scorekey])
                name = dct[namekey].strip()
                ret[lvl][rank] = HighScore(name, score)
        except KeyError:
            return False
        except (TypeError,ValueError):
            raise NHighError('Bad data received')
        _postProcessLevelScores(ret[lvl])
    return True

def _downloadAllEpisodeScores(ep):
    ret = [[emptyHighScore]*20 for i in range(6)]
    #sometimes we get bad data from the server - so try again
    for trynum in xrange(5):
        if _downloadAllEpisodeScoresOnce(ep,ret):
            break
        time.sleep(5)
    else:
        raise NHighError('Error reading scores for episode %d'%ep)
    time.sleep(0.5)
    return ret

class HSDownloader(object):
    def __iter__(self):
        '''Generator that does a partial download and yields
        the amount done (0.0 to 1.0).'''
        self._table = []
        for ep in xrange(NUM_EPISODES):
            yield ep*1.0/NUM_EPISODES
            self._table.append(_downloadAllEpisodeScores(ep))
        yield 1.0
    def result(self):
        '''The resulting HSTable. Only call this when the iterator
        is finished.'''
        return HSTable(self._table)

def fixFileName(fname):
    if '.' not in fname:
        return fname + '.hs'
    return fname

#the pickled data is a tuple of:
#(format_ver (1), date/time, [100*[6*[20*(name, score)]]])

def saveScores(hsTable, fname):
    fname = fixFileName(fname)
    f=file(fname, 'wb')
    lst = [[[(entry.name, entry.score)
             for entry in lvl_entries]
             for lvl_entries in epi_entries]
             for epi_entries in hsTable.table]
    data = (1, hsTable.timestamp, lst)
    cPickle.dump(data, f, 2)
    f.close()

def loadScores(fname):
    try:
        fname = fixFileName(fname)
        f=file(fname, 'rb')
        data = cPickle.load(f)
        f.close()
        if data[0]!=1:
            raise NHighError("Bad file format")
        table = [[[HighScore(name,score)
                   for (name,score) in lvl_entries]
                   for lvl_entries in epi_entries]
                   for epi_entries in data[2]]
        for epi_entries in table:
            for lvl_entries in epi_entries: #lvl_entries = 20* (name, time)
                _postProcessLevelScores(lvl_entries)
        return HSTable(table, data[1])
    except (TypeError,ValueError):
        raise NHighError("Bad file format")

def getPlayerZeroth(hsTable, player):
    '''Find all 0th for a player.
    Return a list of tuples (episode, level, hs entry)'''
    pllower = player.lower()
    ret = []
    for ep_num, ep_entries in enumerate(hsTable.table):
        for lvl_num, lvl_entries in enumerate(ep_entries):
            numOf0th = _getNumOfZeroth(lvl_entries)
            for rank in xrange(numOf0th):
                entry = lvl_entries[rank]
                if entry.name.lower() == pllower:
                    ret.append((ep_num, lvl_num, entry))
                    break
    return ret

def getPlayerScores(hsTable, player):
    '''Find all highscores for a player.
    Return a list of tuples (episode, level, rank, hs entry)'''
    pllower = player.lower()
    ret = []
    for ep_num, ep_entries in enumerate(hsTable.table):
        for lvl_num, lvl_entries in enumerate(ep_entries):
            for rank, entry in enumerate(lvl_entries):
                if entry.name.lower()==pllower:
                    ret.append((ep_num, lvl_num, rank, entry))
                    break
    return ret

#eru_bahagon added the 8 following function
def getPlayerEpScores(hsTable, player):
    '''Find all episode highscores for a player.
    Return a list of tuples (episode, rank, hs entry)'''
    pllower = player.lower()
    ret = []
    for ep_num, ep_entries in enumerate(hsTable.table):
        for rank, entry in enumerate(ep_entries[5]):
            if entry.name.lower()==pllower:
                ret.append((ep_num, rank, entry))
                break
    return ret

def getPlayerNonScores(hsTable, player):
    '''Find all levels/episodes where "player" doesn't 
    have a score '''
    pllower = player.lower()
    ret = []
    for ep_num, ep_entries in enumerate(hsTable.table):
        for lvl_num, lvl_entries in enumerate(ep_entries):
            if pllower not in getAllNamesInEntry(lvl_entries) :
                ret.append((ep_num, lvl_num))
    return ret

def getPlayer1st(hsTable, player):
    '''Find all 1st held by a player and indicates who has the 0th
    ret is {k:0th holder, v:list of 0th he has over player's 1st}'''
    pllower = player.lower()
    ret = {}
    twenty = getPlayerScores(hsTable, player)
    for ep_num, lvl_num, r, entry in twenty :
        if r == 1:
            zeroth = hsTable.table[ep_num][lvl_num][0].name
            if not zeroth in ret.keys():
                ret[zeroth] = []
            ret[zeroth].append([ep_num, lvl_num, entry])
    return ret
        
            

def getAllNamesInEntry(lvl_entries):
    '''Return a list of all player names
    having a score on this specified lvl'''
    ret = []
    for entry in lvl_entries:
        ret.append(entry.name.lower())
    return ret
    
def getLastEntry(lvl_entries):
    '''Return the latest entry of a lvl'''
    scoremin = 10000
    rank = 0
    for r, entry in enumerate(lvl_entries):
        if entry.score == 0:
            break
        if entry.score < scoremin :
            scoremin = entry.score
            rank = r
    return lvl_entries[rank]


def getPlayerWorstLvlScores(hsTable, player):
    '''list the level scores where the diff to 0th is maximal'''
    pllower = player.lower()
    ret = []
    for ep_num, ep_entries in enumerate(hsTable.table):
        for lvl_num, lvl_entries in enumerate(ep_entries[:5]):
            _0th = lvl_entries[0].time
            diff = 0
            for rank, entry in enumerate(lvl_entries):
                if pllower == entry.name.lower() :
                    diff = _0th - entry.time
                    ret.append((ep_num, lvl_num, rank, entry, diff))
                    break
    sortList(ret, reverse = True, key=itemgetter(4))
    return ret

def getPlayerWorstEpScores(hsTable, player):
    '''list the episode of the player where the diff to 0th is maximal'''
    pllower = player.lower()
    ret = []
    for ep_num, ep_entries in enumerate(hsTable.table):
        _0th = ep_entries[5][0].time
        diff = 0
        for rank, entry in enumerate(ep_entries[5]):
            if pllower == entry.name.lower():
                diff = _0th - entry.time
                ret.append((ep_num, rank, entry, diff))
                break
    sortList(ret, reverse = True, key=itemgetter(3))
    return ret

def getPlayerEntry(hsTable, player, ep, lvl):
    '''return the entry, if exists'''
    pllower = player.lower()
    for entry in hsTable.table[ep][lvl]:
        if entry.name.lower() == pllower:
            return entry
    return emptyHighScore
#end of edit    

def findZerothImprovements(hsTableOld, hsTableNew):
    '''Find places where the 0th improved.
    Return list of tuples (episode, level, old entry, new entry)'''
    ret = []
    for episode in xrange(NUM_EPISODES):
        for level in range(6):
            oldZeroth = hsTableOld.table[episode][level][0]
            newZeroth = hsTableNew.table[episode][level][0]
            if oldZeroth.score != newZeroth.score:
                ret.append((episode, level, oldZeroth, newZeroth))
    return ret

def getLevelName(episode, level):
    if level==5:
        return '%02d'%episode
    else:
        return '%02d-%d'%(episode,level)

#add parameters to get smallest differences and get rid of tied scores
def getLevelRankDifferences(hsTable, rank1, rank2, rev=True, tie=True): 
    '''calc differences between 2 different ranks for all levels.
    if ranks are tied, increments rank 2 until first non-tie
      return sorted down list of tuples: (ep_num, lvl_num, diff, entry1, entry2)'''
    assert 0<=rank1<rank2<=19
    ret = []
    for ep_num, ep_entries in enumerate(hsTable.table):
        for lvl_num, lvl_entries in enumerate(ep_entries[:5]):
            #need to be careful because removing fake scores can result in empty entries
            rank1now = rank1
            rank2now = rank2
            while lvl_entries[rank2now]==emptyHighScore and rank2now>0:
                rank2now -= 1
            if rank1now > rank2now:
                rank1now = rank2now
            while not tie and lvl_entries[rank1now].time == lvl_entries[rank2now].time and rank2now < 19 :
                rank2now += 1
            ret.append( (ep_num,lvl_num,
                         lvl_entries[rank1now].time-lvl_entries[rank2now].time,
                         lvl_entries[rank1now], lvl_entries[rank2now]) )
    
    sortList(ret, reverse=rev, key=itemgetter(2))
    return ret

#eru_bahagon's edit	(added 4 following function)
def getEpRankDifferences(hsTable, rank1, rank2, rev=True):
    '''calc differences between 2 different ranks for all episodes.
      return sorted down list of tuples: (ep_num, diff, entry1, entry2)'''
    assert 0<=rank1<rank2<=19
    ret = []
    for ep_num, ep_entries in enumerate(hsTable.table):
		#need to be careful because removing fake scores can result in empty entries
		rank1now = rank1
		rank2now = rank2
		while ep_entries[5][rank2now]==emptyHighScore and rank2now>0:
			rank2now -= 1
		if rank1now > rank2now:
			rank1now = rank2now
		ret.append( (ep_num,
					 ep_entries[5][rank1now].time-ep_entries[5][rank2now].time,
					 ep_entries[5][rank1now], ep_entries[5][rank2now]) )

    sortList(ret, reverse=rev, key=itemgetter(1)) #the key is the difference
    return ret
	
	
def getEpLevelDifference(hsTable, rev=True):
	''' calc differences bewteen all level 0th and ep 0th for all episodes'''
	ret = []
	for ep_num, ep_entries in enumerate(hsTable.table):
		total_lvl_0th = -360
		for lvl_entries in ep_entries[:5]:
			total_lvl_0th += lvl_entries[0].time
		ret.append((ep_num, total_lvl_0th-ep_entries[5][0].time, ep_entries[5][0]))
	sortList(ret, reverse=rev, key=itemgetter(1))
	return ret

def getPoints(hsTable, includeEpisodes=True, includeLevels=True):
    ''' return a dict with player and their number of point
    0th = 20 pts, 1st = 19 pts, etc...'''
    ret = {}
    def _getPlayerEntry(plyr):
        new_entry = [plyr,0]
        if not plyr: return new_entry #don't count empty scores
        return ret.setdefault(plyr.lower(), new_entry) 

    for ep in hsTable.table:
        if not includeEpisodes:
            ep = ep[:5]
        if not includeLevels:
            lvl = ep[5]
            num0th = _getNumOfZeroth(lvl)
            for rank, hsentry in enumerate(lvl):
                plentry = _getPlayerEntry(hsentry.name)
                if rank < num0th:
                    plentry[1] += 20
                else :
                    plentry[1] += (20-rank)
        else :
            for lvl in ep: #each level consists of 20 entries name-time
                num0th = _getNumOfZeroth(lvl)
                for rank, hsentry in enumerate(lvl):
                    plentry = _getPlayerEntry(hsentry.name)
                    if rank < num0th:
                        plentry[1] += 20
                    else :
                        plentry[1] += (20-rank)
    return ret

def getRankDistribution(hsTable, player):
    ''' return the number of times that a player is ranked 0th, 1st, etc...)'''
    ret = [0 for i in range(20)]
    scores = getPlayerScores(hsTable, player)
    for i in scores :
        rank = i[2]
        ret[rank] += 1
    return ret
#end of edit

def _getNumOfZeroth(lvlEntries):
    'Return the number of tied 0th in a level given its entries'
    if not config.allowZerothTies:
        return 1
    zerothScore = lvlEntries[0].score
    for i,entry in enumerate(lvlEntries):
        if entry.score < zerothScore:
            return i
    return 20

#eru_bahagon edit (add includeLevels boolean & modified in consequence)
def getAllPlayerStats(hsTable, includeEpisodes=True, includeLevels=True, low_rank=0, top_rank=9):
    '''returns a dict from lowercase player name
       to [player_name, num_zeroth, num_scores'''
    ret = {}
    def _getPlayerEntry(plyr):
        new_entry = [plyr,0,0,0]
        if not plyr: return new_entry #don't count empty scores
        return ret.setdefault(plyr.lower(), new_entry) 
    for ep in hsTable.table:
        if not includeEpisodes:
            ep = ep[:5]
        if not includeLevels:
            lvl = ep[5]
            num0th = _getNumOfZeroth(lvl)
            for rank, hsentry in enumerate(lvl):
                plentry = _getPlayerEntry(hsentry.name)
                if rank < num0th:
                    plentry[1] += 1
                if low_rank <= rank <= top_rank:
                    plentry[3] += 1 #top 10 rankings
                plentry[2] += 1
        else :
            for lvl in ep: #each level consists of 20 entries name-time
                num0th = _getNumOfZeroth(lvl)
                for rank, hsentry in enumerate(lvl):
                    plentry = _getPlayerEntry(hsentry.name)
                    if rank < num0th:
                        plentry[1] += 1
                    if low_rank <= rank <= top_rank:
                        plentry[3] += 1
                    plentry[2] += 1
    return ret

#eru_bahagon added 3 following function
def getTotalScore(hsTable, get_level=True): #total ep or lvl score
    total = 0
    for ep in hsTable.table:
        if get_level :
            ep = ep[:5]
            for lvl in ep:
                total += lvl[0].time
        else : 
            total += ep[5][0].time
    return total

def getZerothList(hsTable, includeLevels=True, includeEpisodes=True):    
    ret = []
    for ep_num, ep_entries in enumerate(hsTable.table):
        if not includeEpisodes:
            ep_entries = ep_entries[:5]
        if not includeLevels:
            lvl_entry = ep_entries[5][0]
            ret.append((ep_num, 5, lvl_entry))
        else :
            for lvl_num, lvl_entries in enumerate(ep_entries): 
                ret.append((ep_num, lvl_num, lvl_entries[0]))
    return ret
   
 
def getTime(score):
            hours, frames = divmod(score, (60*60))
            mins, frames = divmod(frames, (60))
            secs = frames
            return '%d:%02d:%06.3f' % (hours,mins,secs)

#end of edit

##### replay download/analysis #####

def getReplayKey(ep, lvl, rank):
    if ep<0 or ep>=NUM_EPISODES or lvl<0 or lvl>4 or rank<0 or rank>19:
        return None
    url = 'http://www.harveycartel.org/metanet/n/data13/get_topscores_query_jg.php'
    postdata = 'episode%5Fnumber=' + str(ep)
    allscores = openURL(url, postdata).replace('\r','')
    searchstr = r'&%dpkey%d=(\d+)' % (lvl, rank)
    m=re.search(searchstr, allscores)
    if not m:
        return None
    return int(m.group(1))

def downloadReplayByPKey(pkey):
    url = 'http://www.harveycartel.org/metanet/n/data13/get_lv_demo.php'
    postdata = 'pk='+str(pkey)
    return openURL(url, postdata).replace('\r','')

def downloadReplay(ep, lvl, rank):
    'Returns tuple: (player,score,demo)'
    try:
        pkey = getReplayKey(ep, lvl, rank)
        if pkey is None:
            raise NHighError('Unable to download replay - replay key not found')

        alldata = downloadReplayByPKey(pkey)
        m_name=re.search('&name=([^&]+)',alldata)
        m_score=re.search('&score=([^&]+)',alldata)
        m_demo=re.search('&demo=([^&]+)',alldata)
        try:
            player = m_name.group(1)
            demo = m_demo.group(1)
            score = int(m_score.group(1))
        except (ValueError, AttributeError):
            raise NHighError('Unable to download replay - received invalid data')

        return (player, score, demo)        
        
    except IOError:
        raise NHighError('Error downloading replay data')

def getReplayKeyName(key):
    ret = ''
    if key&1: ret += 'L'
    if key&2: ret += 'R'
    if key&4: ret += 'J'
    if key==0: ret = 'Nothing'
    return ret

def parseReplay(data):
    '''Parse the replay and return array of numbers (1 per frame).
    The numbers can be passed to getReplayKeyName.'''

    if data and data[0]=='$':
        #skip level data
        m = re.match(r'^\$[^#]*#[^#]*#[^#]*#[^#]*#([^#]*)',data)
        if not m:
            raise NHighError('Invalid demo data')
        data = m.group(1)

    m=re.match(r'^(\d+):((\d+\|)*\d+)$', data)
    if not m:
        raise NHighError('Invalid demo data')
    frames = int(m.group(1))
    keystring = m.group(2)
    nums = [int(x) for x in re.findall(r'\d+',keystring)]
    ret = []
    for num in nums:
        numbits = min(frames, 7)
        for bit in xrange(numbits):
            ret.append(num & 0xF)
            num >>= 4
            
        frames -= numbits
        if frames==0 and num!=0:
            #framecount is lower than number of provided frames
            return ret
    return ret

    
def parseAndCompactReplay(data):
    '''Parse the replay, and then find sequences of unchanged keys.
    Return array of tuples (frame_num, key_num, # of occurences)
    The key numbers can be passed to getReplayKeyName.'''

    ret = []
    keys=parseReplay(data)
    if not keys:
        return ret
    keys.append(-1)
    lastkey=-1
    count=0
    frame=0
    for key in keys:
        key = key&7
        if key==lastkey:
            count+=1
        else:
            if lastkey!=-1:
                ret.append((frame-count, lastkey, count))
            lastkey = key
            count = 1
        frame += 1
    return ret

########

def getPlatform():
    for plat in ['win','darwin','linux','mac']:
        if sys.platform.startswith(plat):
            return plat
    return sys.platform

def findPreferencesDir(appname):
    """
    Typical app preference directories are:
        Windows:    C:\Documents and Settings\<username>\Application Data\<owner>\<appname>
        Mac OS X:   ~/Library/Preferences/<appname>
        Unix:       ~/.<lowercased-appname>
    """
    platform = getPlatform()
    if platform=='win':
        try:
            path = os.environ['appdata']
            path = os.path.join(path, appname)
        except KeyError:
            path = None
    elif platform == 'darwin' or platform == 'mac':
        try:
            from Carbon import Folder, Folders
            path = Folder.FSFindFolder(Folders.kUserDomain,
                                       Folders.kPreferencesFolderType,
                                       Folders.kDontCreateFolder)
            path = os.path.join(path.FSRefMakePath(), appname)
        except ImportError:
            path = os.path.expanduser("~/Library/Preferences/"+appname)
    else:
        path = os.path.expanduser("~/." + appname.lower())
    return path


##### .sol handling #####

def findSolFile():
    platform = getPlatform()
    if platform in ['win','darwin','mac']:
        macromediaDir = findPreferencesDir('Macromedia')
        if not macromediaDir:
            return None
        soBaseDir = os.path.join(macromediaDir, 'Flash Player', '#SharedObjects')
        if not os.path.isdir(soBaseDir):
            #I'm not sure if this version actually exists - metanet FAQ mentions it for Mac OS 9
            soBaseDir = os.path.join(macromediaDir, 'Flash Player', '#Shared Objects')
            if not os.path.isdir(soBaseDir):
                return None
        soFiles = [os.path.join(soBaseDir, x) for x in os.listdir(soBaseDir)]
        soDirs = [x for x in soFiles if os.path.isdir(x)]
        if not soDirs:
            return None
        soDir = soDirs[0]
    else: #linux/unix/etc
        soDir = os.path.expanduser('~/.macromedia/Macromedia/Flash Player')

    solFile = os.path.join(soDir, 'localhost', 'n_v14b_userdata.sol')
    if not os.path.exists(solFile):
        return None
    return solFile


class SolReader(object):
    def __init__(self):
        self.readFuncs ={
            0: self.readNumber,
            1: self.readBool,
            2: self.readStr,
            3: self.readObj,
            5: self.readNull,
            6: self.readUndef,
            8: self.readArray,
            #10: self.readRawArr,
            #11: self.readDate,
            #13: self.readObjM,
            #15: self.readObjXML,
            #16: self.readCustomClass,
        }

    def readFromFile(self, size):
        ret = self.f.read(size)
        if len(ret) < size:
            raise EOFError()
        return ret
        
    def readStr(self):
        length, = struct.unpack('>H',self.readFromFile(2))
        if length==0:
            return ''
        s = self.readFromFile(length)
        return s.decode('utf-8')

    def readSol(self, f):
        ret = {}
        try:
            self.f = f
            self.readFromFile(2) #header?
            datasize, = struct.unpack('>L',self.readFromFile(4)) #datasize == filesize-6
            self.readFromFile(4) #filetype=='TCSO' ?
            self.readFromFile(6) #??
            self.readStr() #the .sol name == n_v14b_userdata
            self.readFromFile(4) #??

            while True:
                name = self.readStr()
                val = self.readValue()
                ret[name] = val
                self.readFromFile(1) # == 0
        except EOFError:
            pass
        except (ValueError, TypeError, struct.error):
            raise NHighError('.sol file contains invalid format')
        return ret

    def readNumber(self):
        val, = struct.unpack('>d', self.readFromFile(8))
        return val

    def readBool(self):
        return bool(ord(self.readFromFile(1)))

    def readObj(self):
        ret = {}
        while True:
            name = self.readStr()
            if not name: break
            val = self.readValue()
            ret[name] = val
        self.readFromFile(1) # ==9
        return ret

    def readArray(self):
        length, = struct.unpack('>L',self.readFromFile(4))
        last = -1
        ret = []
        while True:
            name = self.readStr()
            if not name: break
            try:
                now = int(name)
            except ValueError:
                raise NHighError('.sol File contains invalid array object')
            if now != last+1:
                raise NHighError('.sol File contains invalid array object')
            last = now
            ret.append(self.readValue())
        if last != length-1:
            raise NHighError('.sol File contains invalid array object')
        self.readFromFile(1) # ==9
        return ret

    def readNull(self):
        return None

    def readUndef(self):
        return None

    def readValue(self):
        typ = ord(self.readFromFile(1))
        func = self.readFuncs.get(typ)
        if func:
            return func()
        else:
            raise NHighError('.sol File contains unrecognized objects')


def readSolFile():
    filename = findSolFile()
    if not filename:
        raise NHighError('.sol file not found')
    try:
        f = file(filename, 'rb')
        try:
            return SolReader().readSol(f)
        finally:
            f.close()
    except (IOError,OSError):
        raise NHighError('Error reading .sol file')

##############


def calcTotalScores(solData):
    totalEpisodeScore = 0
    totalLevelScore = 0
    try:
        for ep in xrange(NUM_EPISODES):
            epData = solData['persBest'][ep]
            totalEpisodeScore += int(epData['ep']['score'])
            for lvl in xrange(5):
                totalLevelScore += int(epData['lev'][lvl]['score'])
    except (KeyError,TypeError):
        raise NHighError('.sol data is incomplete')
    return (totalLevelScore, totalEpisodeScore)

def findUnsubmittedTop20(solData, hsTable):
    '''Returns: (player, results)
    where results is a list of tuples (ep_num, lvl_num, rank, unsubmitted_score, old_rank, old_entry)
    '''
    ret = []
    player = unicode2str(solData['username'])
    pllower = player.lower()
    try:
        for ep in xrange(NUM_EPISODES):
            epSolData = solData['persBest'][ep]
            solScores = [0]*6
            for lvl in xrange(5):
                solScores[lvl] = int(epSolData['lev'][lvl]['score'])
            solScores[5] = int(epSolData['ep']['score'])
            
            epHSData = hsTable.table[ep]
            for lvl in xrange(6):
                solScore = solScores[lvl]
                lvlHSData = epHSData[lvl]
                for rank,entry in enumerate(lvlHSData):
                    if entry.score == 0:
                        break
                    if entry.score < solScore:
                        oldEntry = None
                        oldRank = None
                        for r, en in enumerate(lvlHSData):
                            if en.name.lower() == pllower:
                                oldEntry = en
                                oldRank = r
                                break
                        ret.append((ep, lvl, rank, solScore, oldRank, oldEntry))
                        break
                    if entry.name.lower() == pllower:
                        break
        return (player, ret)
    except (KeyError,TypeError):
        raise NHighError('.sol data is incomplete')


#eru_bahagon added 18 following function
def findNonTop20Scores(solData, hsTable):
    ''''Returns: (player, results)
    where results is a list of tuples (ep_num, lvl_num, difference to 19th)
    '''
    ret = []
    player = unicode2str(solData['username'])
    pllower = player.lower()
    data = getPlayerNonScores(hsTable, pllower)
    try:
        for ep_num, lvl_num in data:
            epSolData = solData['persBest'][ep_num]
            solScore = 0
            if lvl_num < 5 :
                solScore = int(epSolData['lev'][lvl_num]['score'])
            else : solScore = int(epSolData['ep']['score'])
            
            
            lvlHSData = hsTable.table[ep_num][lvl_num]
            _19th = getLastEntry(lvlHSData)
            
            diff = _19th.score - solScore
            
            ret.append((ep_num, lvl_num, diff))
            
        return (player, ret)
    except (KeyError,TypeError):
        raise NHighError('.sol data is incomplete')

def findEpLvlDifference(solData, hsTable, rev = True):
    ''''Returns: (player, results)
    where results is a list of tuples (ep_num, ep_entry, difference)
    '''
    ret = []
    player = unicode2str(solData['username'])
    pllower = player.lower()
    try:
        for ep in xrange(NUM_EPISODES):
            epSolData = solData['persBest'][ep]
            solScores = [0]*6
            for lvl in xrange(5):
                solScores[lvl] = int(epSolData['lev'][lvl]['score'])
            solScores[5] = int(epSolData['ep']['score'])
            
            epHSData = getPlayerEntry(hsTable, player, ep, 5)
            bestEp = 0
            for lvl in xrange(5):
                solScore = solScores[lvl]
                lvlHSData = getPlayerEntry(hsTable, player, ep, lvl)
                bestEp += max((solScore, lvlHSData.score))
             
            ret.append((ep, epHSData, bestEp*FRAME_TIME-epHSData.time-360))
                
        sortList(ret, reverse=rev, key=itemgetter(2))        
        return (player, ret)
    except (KeyError,TypeError):
        raise NHighError('.sol data is incomplete')
    
    
#eru_bahagon edits : added function to compute total game score (get from other url)
def remove_html_tags_and_replace_them_with_whitespace(data): # does exactly what it says on the tin
        p = re.compile(r'<.*?>')
        return p.sub(' ', data)
    
def remove_extra_spaces(data):
        result = re.compile(r'\s+') # removes extra spaces (when you have tags following each other, 
        #the previous method will put multiple spaces). 
        #Here, we replace every sequence of more than one space by exactly one space
        return result.sub(' ', data)

def lvlToString(ep_num, lvl_num):
    if len(str(ep_num)) < 2 :
        return '0'+str(ep_num)+str(lvl_num)
    else : return str(ep_num)+str(lvl_num)

def stringToLvl(key): 
    nbr = int(key)
    lvl_num = nbr % 10
    ep_num = nbr/10
    return (ep_num, lvl_num)

def isEpKey(key):
    nbr = int(key)
    if nbr%10 == 5:
        return True
    return False

def isLvlKey(key):
    nbr = int(key)
    if nbr%10 == 0 or nbr%10 == 1 or nbr%10 == 2 or nbr%10 == 3 or nbr%10 == 4:
        return True
    return False

def getAllEps(keys):
    ret = []
    for k in keys:
        ep_num = stringToLvl(k)[0]
        if ep_num not in ret :
            ret.append(ep_num)
    sortList(ret)
    return ret
        

def computeTotalScoreRankings(dict):
    #where dict is a dictionnary with players as keys and their corresponding
    #scores dictionnary (key =ep+lvl, value = time)
    ret = {}
    for player in dict.keys():
        scores = dict[player]
        total_ep = 0
        total_lvl = 0
        for k in scores.keys() :
            if isLvlKey(k):
                total_lvl += scores[k]
            if isEpKey(k):
                total_ep += scores[k]
        
        entry = [player, total_lvl, total_ep]
        ret.setdefault(player.lower(), entry)
    return ret

def computePlayerTotalScore(scores):
    #scores is a score dictionnary
    ret = []
    total_ep = 0
    total_lvl = 0
    for k in scores.keys() :
        if isLvlKey(k):
            total_lvl += scores[k]
        if isEpKey(k):
            total_ep += scores[k]
    ret = [total_lvl, total_ep]
    return ret

def computeEpLvlDiff(scores, rev = True):
    ret = []
    epList = getAllEps(scores.keys())
    for ep in epList:
        try :
            epScore = scores[lvlToString(ep,5)]
            bestEp = 0
            for lvl in range(5):
                key = lvlToString(ep, lvl)
                bestEp += scores[key]
                
            ret.append((ep, epScore, bestEp-epScore-360))
        except KeyError:
            pass
    sortList(ret, reverse=rev, key=itemgetter(2)) 
    return ret

def findPossibleEp0th(hsTable, scores, player):
    ret = []
    epList = getAllEps(scores.keys())
    for ep in epList:
        try :
            ep0th = hsTable.table[ep][5][0] #it is an entry, not a score
            bestEp = 0
            for lvl in range(5):
                key = lvlToString(ep, lvl)
                bestEp += scores[key]
            if bestEp-360 > ep0th.time and ep0th.name != player:
                ret.append((ep, ep0th, bestEp-360, bestEp-ep0th.time-360))
        except KeyError:
            pass
    sortList(ret, reverse=True, key=itemgetter(3)) 
    return ret

def computePlayerNonScores(hsTable, scores, player):
    ret = []
    nonList = getPlayerNonScores(hsTable, player)
    for (ep, lvl) in nonList:
        current = 0
        key = lvlToString(ep, lvl)
        if key in scores.keys():
            current = scores[key]
        _19th = getLastEntry(hsTable.table[ep][lvl])
        ret.append((ep, lvl, _19th.time - current))
    return ret

def computePlayerWorstLvlScores(hsTable, scores, player):
    '''list the level scores where the diff to 0th is maximal'''
    pllower = player.lower()
    ret = []
    zeroth = getZerothList(hsTable, includeEpisodes=False)
    for ep, lvl, entry in zeroth:
        current = 0
        key = lvlToString(ep, lvl)
        if key in scores.keys():
            current = scores[key]
        diff = entry.time - current
        ret.append((ep, lvl, current, diff))
    sortList(ret, reverse=True, key=itemgetter(3))
    return ret
        

def getNamesList(hsTable, nbr = 20):
    #returns the first 20 (or nbr) names of regular top 20 rankings
    ret = []
    stats = getAllPlayerStats(hsTable).values()
    if nbr > len(stats):
        nbr = len(stats)
    stats = [(x[2],x[0]) for x in stats if x[2]>0] #x iterates over players (keys)
    sortList(stats, reverse=True)
    for i in xrange(nbr):
        ret.append(stats[i][1])
    return ret

class OnlineScoresDownloader():
    def __init__(self, players):
        self.players = players
    def __iter__(self):
        self.ret = {}
        for i,player in enumerate(self.players):
            yield i*1.0/len(self.players)
            scores = _downloadPlayerOnlineScore(player)
            self.ret.setdefault(player, scores)
        yield 1.0
    def result(self):
        return self.ret
        
def _downloadPlayerOnlineScoreOnce(player, ret):
    #returns all the player's online score, including non-top 20 runs.
    #works like .sol file, for windows/nreality users with corrupted/protected sol file
    #dict with key (ep, lvl) and value = score
    pllower = player.lower()
    url = 'http://www.harveycartel.org/metanet/n/data13/n_highscores_u.php?name='+pllower
    try : 
        data = openURL(url)
    except IOError: 
        return False
    
    temp=remove_extra_spaces(remove_html_tags_and_replace_them_with_whitespace(data)).split()[76:]
        
    if not temp[0].isdigit() :
        raise NHighError("Player %s doesn't exist !" % player)
    temp_ep_score=temp[:temp.index('level')-6] # episode scores
    temp_lvl_score=temp[temp.index('level')+5:temp.index('Note:')] # level scores  
    
    
    for i in range(len(temp_ep_score)/4):
        lvl = 5
        ep = int(temp_ep_score[4*i+2])
        if ep > 99 :
            break
        key = lvlToString(ep, lvl)
        ret.setdefault(key, float(temp_ep_score[4*i+1]))
    
    for i in range(len(temp_lvl_score)/4):
        ep = int(temp_lvl_score[4*i+2])
        if ep > 99 :
            break
        lvl = int(temp_lvl_score[4*i+3])
        key = lvlToString(ep, lvl)
        ret.setdefault(key, float(temp_lvl_score[4*i+1]))
    
    return True

def _downloadPlayerOnlineScore(player):
    ret = {}
    for trynum in xrange(5):
        try : 
            if _downloadPlayerOnlineScoreOnce(player,ret):
                break
        except NHighError as e:
            raise NHighError(str(e))
        time.sleep(5)
    else:
        raise NHighError('Error reading scores for player %s'%player)
    time.sleep(0.5)
    return ret

#end of edit
    
config = NHighConfig()

if __name__=='__main__':
    print 'Do not run this file - run nhigh.py instead'
    raw_input('Press Enter...')
